# Copyright 2020 Newcastle University.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from data_warehouse_client import data_warehouse
import matplotlib.pyplot as pyplot
from matplotlib.backends.backend_pdf import PdfPages
import print_metadata_table
import datetime
import numpy as np


def plot_measurements(dw, rows, study, measurement_type_id, plot_file):
    """
    Plot the value of a measurement over time.
    :param dw: data warehouse handle
    :param rows: a list of measurements generated by the other client functions. Each measurement is in the form:
                    id,time,study,participant,measurementType,
                    typeName,measurementGroup,groupInstance,trial,valType,value
    :param study: study id
    :param measurement_type_id: the measurement type of the measurements to be plotted
    :param plot_file: the name of the file into which the plot will be written
    """
    # https://matplotlib.org/api/pyplot_api.html
    if len(rows) > 0:
        trans = [list(i) for i in zip(*rows)]  # transpose the list of lists
        x = trans[1]  # the data and time
        y = trans[10]  # the measurement value

        mt_info = dw.get_measurement_type_info(study, measurement_type_id)
        units = mt_info[0][3]  # get the units name
        pyplot.title(rows[0][5])
        pyplot.xlabel("Time")  # Set the x-axis label
        pyplot.ylabel(units)  # Set the y-axis label to be the units of the measurement type
        pyplot.plot(x, y)
        pyplot.savefig(plot_file)
        pyplot.close()
    else:
        print("No values to plot\n")


def mk_pdf_report_file_name(f_dir, report_name, time_string):
    return f_dir + report_name + "-" + time_string + ".pdf"


def plot_distributions(dw, study):
    """
    Plot the distribution of all measurent types in a study
    :param dw: data warehouse handle
    :param study: study id
    :return:
    """
    metadata = print_metadata_table.create_measurement_group_info(dw, study)
    measurement_value_index = 10
    file_dir = "reports/"
    timestamp = datetime.datetime.now()  # use the current date and time if none is specified
    time_fname_str = timestamp.strftime('%Y-%m-%dh%Hm%Ms%S')
    fname = mk_pdf_report_file_name(file_dir, "measurement-distributions", time_fname_str)
    pdf = PdfPages(fname)

    # print(f'Measurement Data Distributions for Study {study}\n', file=f)
    for mg_id, mg_data in metadata.items():
        mg_info = mg_data['message_types']
        mg_name = mg_data['name']
        # print(f'\nMeasurement Group {name} (id = {mg_id})\n', file=f)
        for mt_id, mt_info in mg_info.items():
            mt_name = mt_info['name']
            # print(f'\nMeasurement Type {mt_name} (id = {mt_id})\n', file=f)
            ms = dw.get_measurements(study, measurement_group=mg_id, measurement_type=mt_id)  # get all measurements
            if len(ms) > 0:
                trans = [list(i) for i in zip(*ms)]  # transpose the list of lists
                values = trans[measurement_value_index]  # the measurement value
                values = list(filter(lambda val: val is not None, values))  # remove missing values
                # plot distribution based on value type
                value_type = print_metadata_table.valuetype_to_name()[mt_info['valtype']]
                if value_type in ['Integer', 'Real', 'Bounded Integer', 'Bounded Real']:
                    fig = pyplot.figure()
                    pyplot.boxplot([values], patch_artist=True,
                                   labels=[f'{mt_name}'])
                    pyplot.title(f'{mg_name} ({mg_id}) / {mt_name} ({mt_id})')
                    pyplot.close()
                    pdf.savefig(fig)
                elif value_type in ['Nominal', 'Ordinal']:
                    categories = mt_info['categories']  # get the category dictionary k = name, val = id
                    n_cat = len(categories)  # calculate the number of categories
                    cat_count = np.zeros((n_cat,), dtype=int)  # initialise a vector for each category
                    cat_index = {}
                    index = 0
                    for c in categories:  # create a mapping from the category name to the index into the vector
                        cat_index[c] = index
                        index += 1
                    #  print(f'\n {mg_id} / {mt_id}: {cat_index} \n')
                    for v in values:   # count how many values in each category
                        #  print(*values, sep='\n')
                        cat_count[cat_index[v]] += 1
                    cat_names = list(categories.keys())
                    cat_counts = list(cat_count)
                    fig = pyplot.figure()
                    pyplot.bar(cat_names, cat_counts)
                    pyplot.title(f'{mg_name} ({mg_id}) / {mt_name} ({mt_id})')
                    pyplot.close()
                    pdf.savefig(fig)
                elif value_type in ['Boolean']:
                    t_count = 0
                    f_count = 0
                    for v in values:
                        if v == 'T':
                            t_count += 1
                        else:
                            f_count += 1
                    fig = pyplot.figure()
                    pyplot.bar(['False', 'True'], [f_count, t_count])
                    pyplot.title(f'{mg_name} ({mg_id}) / {mt_name} ({mt_id})')
                    pyplot.close()
                    pdf.savefig(fig)
#              else:
                    # print('No Measurements Recorded', file=f)
    pdf.close()


# test it out
#  dw_handle = data_warehouse.DataWarehouse("db-credentials-read-only.json", "datawarehouse")
#
#  study_id = 8
#
#  plot_distributions(dw_handle, study_id)



# Copyright 2020 Newcastle University.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# Summarise a study

from tabulate import tabulate
from data_warehouse_client import print_io
from data_warehouse_client import csv_io
import datetime


def get_instances_per_measurement_group(dw, study):
    """
    Returns the number of instances in all measurement groups in a study
    :param dw: data warehouse handle
    :param study: study id
    :return: number of instances in each measurement group
    """
    measurement_groups = dw.get_all_measurement_groups(study)
    summary_table = []
    for [mg_id, mg_name] in measurement_groups:
        n_mgi = dw.n_mg_instances(mg_id, study)
        summary_table = summary_table + [(mg_id, mg_name, n_mgi)]
    return summary_table


def print_all_instances_in_a_study(dw, study):
    """
    Print all instances in a study - don't print for measurement groups that have no measurements
    Now depreciated. Use print_instances_in_a_study_to_csv_files
    :param dw: data warehouse handle
    :param study: study id
    """
    print(f'All Measurement Group Instances in Study {study}')
    measurement_groups = dw.get_all_measurement_groups(study)
    print()
    for [mg_id, mg_name] in measurement_groups:
        (header, instances) = dw.get_measurement_group_instances(study, mg_id, [])
        if len(instances) > 0:
            print(f'All measurements in group {mg_id} ({mg_name}) for Study {study} \n')
            print_io.print_measurement_group_instances(header, instances)
            print()


def print_study_summary(dw, study):
    """
    Print a summary of the number of participants and the total instances in each measgeurement group in a study
    :param dw: data warehouse handle
    :param study: study id
    :return:
    """
    print(f'Summary of Study {study}')

    # Get Number of Participants
    n_participants = len(dw.get_participants(study))
    print(f'Number of Participants in Study {study}: {n_participants}')

    # Measurement Groups per Measurement Group Type
    print(f'Total Instances in each Measurement Group in Study {study} that contain measurements:')
    print("\n")

    summary = get_instances_per_measurement_group(dw, study)
    print(tabulate(summary, headers=['Id', 'Name', '#Instances']))

    n_measurement_group_instances = 0
    for (mg_id, mg_name, n_mgi) in summary:
        n_measurement_group_instances = n_measurement_group_instances + n_mgi

    print()
    print(f'Total Measurement Group Instances in Study {study}: {n_measurement_group_instances}')
    print()


def mk_txt_report_file_name(f_dir, report_name, time_string):
    return f_dir + report_name + time_string + ".txt"


def mk_csv_report_file_name(f_dir, report_name, time_string):
    return f_dir + report_name + time_string + ".csv"


def print_all_instances_in_a_study_to_file(dw, study):
    """
    Print to a file all instances in a study - don't print for measurement groups that have no measurements
    :param dw: data warehouse handle
    :param study: study id
    """
    file_dir = "reports/"
    timestamp = datetime.datetime.now()  # use the current date and time if none is specified
    time_fname_str = timestamp.strftime('%Y-%m-%dh%Hm%Ms%S')
    fname = mk_txt_report_file_name(file_dir, "study-instances-", time_fname_str)
    with open(fname, "w", encoding="utf-8") as f:
        print(f'All Measurement Group Instances in Study {study}\n', file=f)
        measurement_groups = dw.get_all_measurement_groups(study)
        for [mg_id, mg_name] in measurement_groups:
            (header, instances) = dw.get_measurement_group_instances(study, mg_id, [])
            if len(instances) > 0:
                print(f'All measurements in group {mg_id} ({mg_name}) for Study {study} \n', file=f)
                print_io.print_measurement_group_instances_to_file(header, instances, f)
                print('\n', file=f)


def print_all_instances_in_a_study_to_csv_files(dw, study):
    """
    Print all instances in a study to a set of csvs - one per measurement group
    Now depreciated. Use print_instances_in_a_study_to_csv_files
    :param dw: data warehouse handle
    :param study: study id
    """
    file_dir = "reports/"
    timestamp = datetime.datetime.now()  # use the current date and time if none is specified
    time_fname_str = timestamp.strftime('%Y-%m-%dh%Hm%Ms%S')
    measurement_groups = dw.get_all_measurement_groups(study)
    for [mg_id, mg_name] in measurement_groups:
        (header, instances) = dw.get_measurement_group_instances(study, mg_id, [])
        if len(instances) > 0:
            fname = mk_csv_report_file_name(file_dir, "study-instances-" + mg_name + "-", time_fname_str)
            csv_io.export_measurement_groups_as_csv(header, instances, fname)


def mk_participants_dictionary(dw, study):
    """
    Create a dictionary containing mapping from participant_id to local_id for all the participants in a study
    :param dw: data warehouse handle
    :param study: study_id
    :return: the dictionary
    """
    participants = dw.get_participants(study)  # return list of ids and local ids
    participant_local_id = {}
    for p in participants:
        participant_local_id[p[0]] = p[1]
    return participant_local_id


def print_instances_in_a_study_to_csv_files(dw, study, report_dir, select_participants=False, participants=[],
                                            local_participant_id=True, filename_prefix='', print_empty_files=False):
    """
    Print instances in a study to a set of csvs - one per measurement group,
    but only for those participants in the participants list
    :param dw: data warehouse handle
    :param study: study id
    :param report_dir: the directory in which the profiles will be written
    :param select_participants: select a subset of participants to be included in the profile
    :param participants: list of participants to be included in the profile if select_participants is true
    :param local_participant_id: include the local participant id if this boolean is True
    :param filename_prefix: optional string to add to front of filename
    :param print_empty_files: optional boolean to print csv files with no instances in them
    """
    if local_participant_id:
        participant_local_id = mk_participants_dictionary(dw, study)   # create a dictionary mapping from id to local id
    timestamp = datetime.datetime.now()  # use the current date and time if none is specified
    time_fname_str = timestamp.strftime('%Y-%m-%dh%Hm%Ms%S')
    participant_index = 3  # the participant_id is in position 3 of the list
    measurement_groups = dw.get_all_measurement_groups(study)
    for [mg_id, mg_name] in measurement_groups:
        (header, all_instances) = dw.get_measurement_group_instances(study, mg_id, [])
        if select_participants:  # select participants
            instances = list(filter(lambda instance: instance[participant_index] in participants, all_instances))
        else:
            instances = all_instances
        if (len(instances) > 0) or print_empty_files:   # if there are some instances in the measurement group
                                                        #  or if files should be created even if there are no instances
            fname = mk_csv_report_file_name(report_dir, filename_prefix +
                                            "study-instances-" + mg_name + "-", time_fname_str)
            if local_participant_id:
                instances_with_local_participant_id = []
                extended_header = ['Local Participant'] + header
                for instance in instances:  # add the local participant id to the start of each row
                    participant_id = instance[participant_index]  # get unique participant id
                    local_participant = participant_local_id[participant_id]  # get the local participant id
                    instances_with_local_participant_id = instances_with_local_participant_id +\
                                                          [[local_participant] + instance]
                    csv_io.export_measurement_groups_as_csv(extended_header, instances_with_local_participant_id, fname)
            else:  # don't include local id
                csv_io.export_measurement_groups_as_csv(header, instances, fname)


def print_all_instances_in_a_study_with_local_participant_id_to_csv_files(dw, study):
    """
    Print all instances in a study to a set of csvs - one per measurement group - including participants
    This is now depreciated. Use print_instances_in_a_study_to_csv_files
    :param dw: data warehouse handle
    :param study: study id
    """
    file_dir = "reports/"
    participant_id_index = 3
    timestamp = datetime.datetime.now()  # use the current date and time if none is specified
    time_fname_str = timestamp.strftime('%Y-%m-%dh%Hm%Ms%S')
    measurement_groups = dw.get_all_measurement_groups(study)
    for [mg_id, mg_name] in measurement_groups:
        (header, instances) = dw.get_measurement_group_instances(study, mg_id, [])
        extended_header = ['Local Participant'] + header
        if len(instances) > 0:  # if there are some instances in the measurement group
            fname = mk_csv_report_file_name(file_dir, "study-instances-" + mg_name + "-", time_fname_str)
            instances_with_local_participant_id = []
            for instance in instances:  # add the local participant id to the start of each row
                participant_id = instance[participant_id_index]  # get unique participant id
                (success, part) = dw.get_participant_by_id(study, participant_id)  # get local participant id
                instances_with_local_participant_id = instances_with_local_participant_id + [[part] + instance]
            csv_io.export_measurement_groups_as_csv(extended_header, instances_with_local_participant_id, fname)


def print_study_summary_to_file(dw, study):
    """
    Print to a file a summary of the number of participants and the total instances in each measurement group in a study
    :param dw: data warehouse handle
    :param study: study id
    :return:
    """
    file_dir = "reports/"
    timestamp = datetime.datetime.now()  # use the current date and time if none is specified
    time_fname_str = timestamp.strftime('%Y-%m-%dh%Hm%Ms%S')
    fname = mk_txt_report_file_name(file_dir, "study-summary-", time_fname_str)
    with open(fname, "w", encoding="utf-8") as f:
        print(f'Summary of Study {study}', file=f)
        # Get Number of Participants
        n_participants = len(dw.get_participants(study))
        print(f'Number of Participants in Study {study}: {n_participants}', file=f)
        # Measurement Groups per Measurement Group Type
        print(f'Total Instances in each Measurement Group in Study {study} that contain measurements:\n', file=f)
        summary = get_instances_per_measurement_group(dw, study)
        print(tabulate(summary, headers=['Id', 'Name', '#Instances']), file=f)

        n_measurement_group_instances = 0
        for (mg_id, mg_name, n_mgi) in summary:
            n_measurement_group_instances = n_measurement_group_instances + n_mgi
        print(f'Total Measurement Group Instances in Study {study}: {n_measurement_group_instances}\n', file=f)



# Copyright 2020 Newcastle University.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import pandas as pd
import pandas_profiling as pp
import datetime


def mk_html_report_file_name(f_dir, report_name, time_string):
    return f_dir + report_name + time_string + ".html"


def profile_all_measurement_groups(dw, study, report_dir, select_participants=False, participants=[],
                                   select_trials=False, trials=[], hide_trial_column=False, filename_prefix=''):
    """
    Write a Profile file for each measurement group in a study
    :param dw: data warehouse handle
    :param study: study id
    :param report_dir: the directory in which the profiles will be written
    :param select_participants: select a subset of participants to be included in the profile
    :param participants: list of participants to be included in the profile if select_participants is true
    :param select_trials: select a subset of trials to be included in the profile
    :param trials: list of trials to be included in the profile if select_trials is true
    :param hide_trial_column: don't include the Trial column (useful for studies where trial is not used)
    :param filename_prefix: optional string to add to front of filename
    """
    timestamp = datetime.datetime.now()                        # use the current date and time in the filenames
    time_fname_str = timestamp.strftime('%Y-%m-%dh%Hm%Ms%S')   # format the date time string used in the filename
    measurement_groups = dw.get_all_measurement_groups(study)  # get all the measurement groups in the study

    for [mg_id, mg_name] in measurement_groups:                # for each measurement group
        (header, instances) = dw.get_measurement_group_instances(study, mg_id, [])   # extract header and all instances
        df_full = pd.DataFrame(instances, columns=header)                            # create a pandas data frame
        if select_participants:   # select participants
            df_participants = df_full.loc[df_full['Participant'].isin(participants)]
        else:
            df_participants = df_full   # include all participants
        if select_trials:  # select trials
            df = df_participants.loc[df_participants['Trial'].isin(trials)]
        else:
            df = df_participants   # include all trials
        # drop columns that are not needed in the profile
        if hide_trial_column:
            std_columns_to_drop = ['Trial', 'Instance', 'Study', 'Measurement Group', 'Participant']
        else:
            std_columns_to_drop = ['Instance', 'Study', 'Measurement Group', 'Participant']
        if 'uniqueId' in header:
            columns_to_drop = ['uniqueId'] + std_columns_to_drop
        else:
            if 'uniqueid' in header:
                columns_to_drop = ['uniqueid'] + std_columns_to_drop
            else:
                columns_to_drop = std_columns_to_drop
        df_main_columns = df.drop(columns=columns_to_drop)
        # create a profile report
        profile = pp.ProfileReport(df_main_columns, title="Profiling Report for " + mg_name, progress_bar=False)
        f_name = mk_html_report_file_name(report_dir,
                                          filename_prefix + "Measurement-Group-Profile-" + mg_name + "-",
                                          time_fname_str)
        profile.to_file(f_name)  # store the report in  a file



# Copyright 2020 Newcastle University.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# only run this if you're certain that you want to remove all metdata and measurements from a study!
from data_warehouse_client import delete_study_contents
from data_warehouse_client import table_writer_json
from data_warehouse_client import data_warehouse

# Create a connection to the data warehouse
data_warehouse = data_warehouse.DataWarehouse("db-credentials.json", "datawarehouse")

study = 86
file_name = "output/metadatatables" + str(study) + ".json"
# store everything important just in case!
# table_writer_json.data_warehouse_metadata_tables_to_file(data_warehouse, study, file_name)


# delete_study_contents.delete_study_contents(data_warehouse, study)

# for study in range(8,26):
# delete_study_contents.delete_study_completely(data_warehouse, ???)



# Copyright 2020 Newcastle University.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest import TestCase

from data_warehouse_client import transform_result_format
import datetime


class TestTransformResultFormat(TestCase):
    def test_form_measurements(self):
        rid = 101
        time = datetime.datetime.now()
        study = 102
        trial = 103
        measurement_group = 104
        group_inst = 99
        measurement_type = 105
        measurement_name = "Measurement Name"
        participant = 106
        int_type = 0
        real_type = 1
        val_type = int_type
        val_integer = 108
        val_real = 109.00
        val_text = "Text Value"
        val_datetime = time
        category_name = "A Category"
        ipr1 = [rid, time, study, participant, measurement_type, measurement_name, measurement_group, group_inst, trial,
                int_type, val_integer, val_real, val_text, val_datetime, category_name]
        ipr2 = [rid, time, study, participant, measurement_type, measurement_name, measurement_group, group_inst, trial,
                real_type, val_integer, val_real, val_text, val_datetime, category_name]
        ip = [ipr1, ipr2]
        opr1 = [rid, time, study, participant, measurement_type, measurement_name, measurement_group, group_inst,
                trial, int_type, val_integer]
        opr2 = [rid, time, study, participant, measurement_type, measurement_name, measurement_group, group_inst,
                trial, real_type, val_real]
        op = [opr1, opr2]
        self.assertEqual(op, transform_result_format.form_measurements(ip))



# Copyright 2020 Newcastle University.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os


def table_to_dictionary(dw, table_name, study_id):
    """
    Read a table and turn it into a dictionary
    Based on code in: https://stackoverflow.com/questions/3286525/return-sql-table-as-json-in-python
    :param dw: data warehouse handle
    :param table_name: Data Warehouse table name
    :param study_id: the study id
    :result a list of dictionaries - one for each row of the table
    """
    cur = dw.dbConnection.cursor()
    q = "select * from " + table_name + " where study = " + str(study_id) + ";"
    cur.execute(q)
    r = [dict((cur.description[i][0], value)
              for i, value in enumerate(row)) for row in cur.fetchall()]
    return r


def tables_to_dictionary(dw, study_id, tables):
    """
    Transform a set of tables into a dictionary
    :param dw: data warehouse handle
    :param study_id: the study id
    :param tables: a list of tables to convert to a dictionary
    :result a dictionary of tables, each a list of dictionary - one for each row of the table
    """
    table_dict = {}
    for table_name in tables:
        td = table_to_dictionary(dw, table_name, study_id)
        table_dict[table_name] = td
    return table_dict


def data_warehouse_metadata_tables_to_dictionary(dw, study_id):
    """
    Writes the data warehouse metadata tables into a dictionary
    :param dw: data warehouse handle
    :param study_id: the study id
    :return a dictionary of tables, each a list of dictionaries - one for each row of the table
    """
    return tables_to_dictionary(dw, study_id, metadata_table_names())


def metadata_table_names():
    """
    :return: the names of all the metadata tables in the data warehouse
    """
    return ["trial", "units", "measurementtype", "measurementgroup", "sourcetype", "boundsreal",
            "boundsint", "category", "measurementtypetogroup"]


def dictionary_to_json_file(d, file_name):
    """
    write dictionary to file in json format
    :param d: dictionary
    :param file_name: output filename
    """
    os.makedirs(os.path.dirname(file_name), exist_ok=True)
    with open(file_name, 'w') as json_file:
        json.dump(d, json_file)


def data_warehouse_metadata_tables_to_file(dw, study_id, file_name):
    """
    Write a set of data warehouse tables into a file
    :param dw: data warehouse handle
    :param study_id: the study id
    :param file_name: output filename
    """
    d = data_warehouse_metadata_tables_to_dictionary(dw, study_id)
    dictionary_to_json_file(d, file_name)


def write_tables_in_dw_from_dictionary(dw, input_dict):
    """
    writes a set of tables (in a dictionary) into the data warehouse
    :param dw: data warehouse handle
    :param input_dict: the dictionary containing the tables
    :return the total number of rows inserted
    """
    total_rows_inserted = 0
    for table in input_dict.keys():
        total_rows_inserted += insert_rows_in_dw_from_dictionary(dw, table, input_dict[table])
    return total_rows_inserted


def insert_rows_in_dw_from_dictionary(dw, table, rows):
    """
    writes a table, in the form of a list of dictionaries (one per row), into the data warehouse
    :param dw: data warehouse handle
    :param table: table name
    :param rows: a list of dictionaries - one per row
    :return number of rows inserted
    """
    # https://blog.softhints.com/python-3-convert-dictionary-to-sql-insert/
    cur = dw.dbConnection.cursor()
    n_rows_inserted = 0
    for row in rows:
        placeholders = ', '.join(['%s'] * len(row))
        columns = ', '.join(row.keys())
        sql = "INSERT INTO %s ( %s ) VALUES ( %s );" % (table, columns, placeholders)
        cur.execute(sql, list(row.values()))
        dw.dbConnection.commit()
        n_rows_inserted += 1
    return n_rows_inserted



# Copyright 2020 Newcastle University.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from data_warehouse_client import file_utils
from tabulate import tabulate  # https://github.com/astanin/python-tabulate
import datetime


def valuetype_to_name():
    return {0: 'Integer', 1: 'Real', 2: 'Text', 3: 'Date Time', 4: 'Boolean',
            5: 'Nominal', 6: 'Ordinal', 7: 'Bounded Integer', 8: 'Bounded Real'}


def mk_txt_report_file_name(f_dir, report_name, time_string):
    return f_dir + report_name + time_string + ".txt"


def print_metadata_tables_to_file(dw, study_id):

    file_dir = "reports/"
    timestamp = datetime.datetime.now()  # use the current date and time if none is specified
    time_fname_str = timestamp.strftime('%Y-%m-%dh%Hm%Ms%S')
    fname = mk_txt_report_file_name(file_dir, "metadata-table-", time_fname_str)

    metadata = create_measurement_group_info(dw, study_id)
    headers = ['Measurement Type', 'Id', 'Value Type', 'Optional?', 'Units', 'Min Value', 'Max Value', 'Categories']

    with open(fname, "w", encoding="utf-8") as f:
        print(f'Measurement Groups for Study {study_id}', file=f)
        for mg_id, mg_data in metadata.items():
            mg_info = mg_data['message_types']
            name = mg_data['name']
            rows = []
            for mt_id, mt_info in mg_info.items():
                cats = '\n'.join(f'{v}:{k}' for k, v in mt_info['categories'].items())
                lower_bound = ''
                upper_bound = ''
                if mt_info['intbounds'] != {}:
                    lower_bound = mt_info['intbounds']['minval']
                    upper_bound = mt_info['intbounds']['maxval']
                elif mt_info['realbounds'] != {}:
                    lower_bound = mt_info['realbounds']['minval']
                    upper_bound = mt_info['realbounds']['maxval']
                if mt_info['units'] is None:
                    units_prnt = ""
                else:
                    units_prnt = mt_info['units']
                row = [mt_info['name'], mt_id, valuetype_to_name()[mt_info['valtype']],
                       mt_info['optional'], units_prnt, lower_bound, upper_bound, cats]
                rows = rows + [row]
            print(f'\nMeasurement Group {name} (id = {mg_id})\n', file=f)
            print(tabulate(rows, headers=headers, tablefmt="grid"), file=f)


def print_metadata_tables(dw, study_id):
    """
    print study metadata
    """
    metadata = create_measurement_group_info(dw, study_id)
    print(f'Measurement Groups for Study {study_id}')
    headers = ['Measurement Type', 'Id', 'Value Type', 'Optional?', 'Units', 'Min Value', 'Max Value', 'Categories']
    for mg_id, mg_data in metadata.items():
        mg_info = mg_data['message_types']
        name = mg_data['name']
        rows = []
        for mt_id, mt_info in mg_info.items():
            cats = '\n'.join(f'{v}:{k}' for k, v in mt_info['categories'].items())
            lower_bound = ''
            upper_bound = ''
            if mt_info['intbounds'] != {}:
                lower_bound = mt_info['intbounds']['minval']
                upper_bound = mt_info['intbounds']['maxval']
            elif mt_info['realbounds'] != {}:
                lower_bound = mt_info['realbounds']['minval']
                upper_bound = mt_info['realbounds']['maxval']
            if mt_info['units'] is None:
                units_prnt = ""
            else:
                units_prnt = mt_info['units']
            row = [mt_info['name'], mt_id, valuetype_to_name()[mt_info['valtype']],
                   mt_info['optional'], units_prnt, lower_bound, upper_bound, cats]
            rows = rows + [row]
        print(f'\nMeasurement Group {name} (id = {mg_id})\n')
        print(tabulate(rows, headers=headers, tablefmt="grid"))


def mk_optional(opt):
    """
    turns the value in the optional column into a boolean
    :param opt: either true, false or None
    :return: True (if opt == true) or False (if Null, or false)
    """
    if opt != True:
        return False
    else:
        return True


def create_measurement_group_info(dw, study):
    """
    Creates a dictionary with each entry a measurement group
    :param dw: data warehouse handle
    :param study: study id
    :return:
    """
    q1 = file_utils.process_sql_template("get_measurement_group_info.sql", {"study": study})
    r1 = dw.return_query_result(q1)  # return a list of (measurementgroup, measurementtype, name, valtype, optional)

    q2 = file_utils.process_sql_template("get_categories_in_study.sql", {"study": study})
    r2 = dw.return_query_result(q2)

    cats = {}
    for mt in set([row[0] for row in r2]):
        cats[mt] = dict(map(lambda t: (t[2], t[1]),  filter(lambda r: r[0] == mt, r2)))

    q3 = file_utils.process_sql_template("get_boundsint_in_study.sql", {"study": study})
    r3 = dw.return_query_result(q3)
    int_bounds = {}
    for [mt_id, minval, maxval] in r3:
        int_bounds[mt_id] = {'minval': minval, 'maxval': maxval}

    q4 = file_utils.process_sql_template("get_boundsreal_in_study.sql", {"study": study})
    r4 = dw.return_query_result(q4)
    real_bounds = {}
    for [mt_id, minval, maxval] in r4:
        real_bounds[mt_id] = {'minval': minval, 'maxval': maxval}

    q5 = file_utils.process_sql_template("get_measurement_groups_in_study.sql", {"study": study})
    r5 = dw.return_query_result(q5)
    mg_names = {}
    for [mg_id, mg_name] in r5:
        mg_names[mg_id] = mg_name

    q6 = file_utils.process_sql_template("get_units_in_study.sql", {"study": study})
    r6 = dw.return_query_result(q6)
    unit_name = {}
    for [mt_id, unit_description] in r6:
        unit_name[mt_id] = unit_description

    mg_info = {}
    for mg in set([row[0] for row in r1]):
        mt_info_all = {}
        for (group, ms_type, name, valtype, optional, units) in list(filter(lambda r: r[0] == mg, r1)):
            cat = cats.get(ms_type)
            if cat is None:
                cat_val = {}
            else:
                cat_val = cat
            intbounds = int_bounds.get(ms_type)
            if intbounds is None:
                int_bounds_val = {}
            else:
                int_bounds_val = intbounds
            realbounds = real_bounds.get(ms_type)
            if realbounds is None:
                real_bounds_val = {}
            else:
                real_bounds_val = realbounds
            unitname = unit_name.get(ms_type)
            if unitname is None:
                units_val = None
            else:
                units_val = unitname
            mt_info = {'name': name, 'valtype': valtype, 'optional': mk_optional(optional), 'units': units_val,
                       'categories': cat_val, 'intbounds': int_bounds_val, 'realbounds': real_bounds_val}
            mt_info_all[ms_type] = mt_info
        mg_info[mg] = {'name': mg_names[mg], 'message_types': mt_info_all}
    return mg_info



# Copyright 2020 Newcastle University.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import functools
from datetime import datetime
import unidecode


def process_message_group(mg_triples):
    """
    takes the result of attempting to load each field in a message group and processes it
    :param mg_triples: list of (successful load?, ((measurement_ttpe, valtype, value)))
    :return: Success?, the list of (measurement_type, valtype, value) triples, the error message list
    """
    success_index = 0
    triple_index = 1
    error_message_index = 2
    oks = list(map(lambda r: r[success_index], mg_triples))     # get a list of the bools indicating success
    triples = list(map(lambda r: r[triple_index], mg_triples))  # get a list of the triples
    if functools.reduce(lambda x, y: x and y, oks):             # if all value in message group are correct....
        return True, sum(triples, []), []                       # return Success, the triples, empty error message list
    else:                                                       # there was a problem with >0 fields in message group
        # remove empty strings from error messages
        error_messages = list(filter(lambda s: s != "", list(map(lambda r: r[error_message_index], mg_triples))))
        return False, [], error_messages  # return Failure, no triples and the list of error messages


def missing_mandatory_type_error_message(jfield, measurement_type, data):
    """
    form error message for missing mandatory types
    :param jfield: the field name
    :param measurement_type: the measurement type
    :param data: the data in which the field is missing
    :return: error message
    """
    return f'Missing mandatory field {jfield} (measurement type {measurement_type}) in data: {data}'


def wrong_type_error_message(jfield, measurement_type, data, val_type):
    """
    form error message for missing mandatory types
    :param jfield: the field name
    :param measurement_type: the measurement type
    :param data: the data in which the field is missing
    :param val_type: the correct type of the field
    :return: error message
    """
    return f'Wrong type for {jfield} (measurement type {measurement_type}) in data {data};' +\
           f'it should be a {type_names(val_type)} (value type {val_type})'


def type_names(val_type):
    """
    return the name of a type represented by a number
    :param val_type: int representing a type in the warehouse
    :return: name
    """
    if val_type in [0, 5, 6, 7]:  # 5 and 6 are there for ordinals and nominals created from id
        return "integer"
    elif val_type in [1, 8]:
        return "real"
    elif val_type == 3:
        return "datetime"
    elif val_type == 4:
        return "boolean"
    else:
        return "string"


def convert_epoch_in_ms_to_string(timestamp_in_ms):
    """
    converts timestamp in ms epoch format to string
    :param timestamp_in_ms: epoch timestamp in ms
    :return: well_formed, date time string
    """
    if check_int(timestamp_in_ms):
        timestamp_in_sec = int(timestamp_in_ms)//1000
        try:
            time_val = datetime.fromtimestamp(timestamp_in_sec)
            return True, str(time_val)
        except OSError as error:
            return False, None
    else:
        return False, None


def convert_posix_timestamp_to_string(val):
    """
    converts timestamp in ms epoch format to string
    :param val: posix timestamp
    :return: well_formed, date time string
    """
    return True, val.replace('T', ' ')


def check_int(string):
    """
    Check if a string represents an integer
    :param string: string
    :return: True if the string represents an integer
    """
    try:
        val = int(string)
        return True
    except ValueError:
        return False


def check_real(string):
    """
    Check if a string represents a real
    :param string: string
    :return: True if the string represents a real
    """
    try:
        val = float(string)
        return True
    except ValueError:
        return False


def check_datetime(string):
    """
    Check if a string represents a datetime
    :param string: string
    :return: True if the string represents a datetime
    """
    return True  # need to add checking later


def check_boolean(string):
    """
    Check if a string represents a boolean
    :param string: 
    :return: True if the string represents a boolean
    """
    return string in ['T', 'Y', 'F', 'N', '0', '1', 0, 1]


def type_check(val, val_type):
    """
    Check the type of a value retrieved from a field
    :param val: value
    :param val_type: the type is should be
    :return: True if the value has the right type, False otherwise
    """
    if val_type in [0, 5, 6, 7]:  # 5 and 6 are there for ordinals and nominals created from id
        well_typed = check_int(val)
    elif val_type in [1, 8]:
        well_typed = check_real(val)
    elif val_type == 3:
        well_typed = check_datetime(val)
    elif val_type == 4:
        well_typed = check_boolean(val)
    else:
        well_typed = True  # everything else is a string
    return well_typed


def get_and_check_value(measurement_type, val_type, data, jfield, optional):
    """
    check if a value exists, and if so its type
    :param measurement_type: measurement type of jfield in the data warehouse
    :param val_type: the type of the value to be stored
    :param data: json that contains the jfield
    :param jfield: the name of the field
    :param optional: if the field is optional
    :return: (field exists,well typed, value, error_message)
    """
    error_message = ""
    val = data.get(jfield)
    if val is None:
        exists = False
        well_typed = False  # default
    elif val == "":
        exists = False
        well_typed = False  # default
    else:
        exists = True
        well_typed = type_check(val, val_type)
    if (not optional) and (not exists):
        error_message = missing_mandatory_type_error_message(jfield, measurement_type, data)
    elif exists and (not well_typed):
        error_message = wrong_type_error_message(jfield, measurement_type, data, val_type)
    return exists, well_typed, val, error_message


def mk_basic_field(measurement_type, val_type, data, jfield):
    """
    Make triple (measurement_ttpe, valtype, value for the jfield in the data)
    :param measurement_type: measurement type of jfield in the data warehouse
    :param val_type: the type of the value to be stored
    :param data: json that contains the jfield
    :param jfield: the name of the field
    :return: Error free, (measurement_type, valtype, value for the jfield in the data), error_message
    """
    (exists, well_formed, val, error_message) = get_and_check_value(measurement_type, val_type, data, jfield, False)
    if exists and well_formed:
        return True, [(measurement_type, val_type, val)], ""
    else:
        return False, [], error_message


def mk_optional_basic_field(measurement_type, val_type, data, jfield):
    """
    If the jfield exists in the data then return [(measurement_ttpe, valtype, value for the jfield in the data)].
    If not then return an empty list.
    :param measurement_type:    measurement type of jfield in the data warehouse
    :param val_type: the type of the value to be stored
    :param data:                json that may contain the jfield
    :param jfield:              the name of the field
    :return                     (Error free, if the field exists then a list is returned holding the appropriate entry;
                                if the field doesn't exist then an empty list is returned, error message)
    """
    (exists, well_formed, val, error_message) = get_and_check_value(measurement_type, val_type, data, jfield, True)
    if exists and well_formed:
        return True, [(measurement_type, val_type, val)], ""  # jfield is present in the json and no errors
    elif exists and not well_formed:  # jfield is present but there are errors
        return False, [], error_message
    else:
        return True, [], ""  # Field doesn't exist, which is OK as this is an optional field


def mk_int(measurement_type, data, jfield):
    """
    create a (measurement_type, valtype, value for the jfield in the data) triple for an integer
    :param measurement_type: measurement type of jfield in the data warehouse
    :param data: json that may contain the jfield
    :param jfield: the name of the field
    :return: Error free, [(measurement_type, valtype, value for the jfield in the data)], error_message
    """
    return mk_basic_field(measurement_type, 0, data, jfield)


def mk_optional_int(measurement_type, data, jfield):
    """
    If the jfield exists in the data then return [(measurement_ttpe, valtype, value for the jfield in the data)].
    If not then return an empty list.
        :param measurement_type:    measurement type of jfield in the data warehouse
        :param data:                json that may contain the jfield
        :param jfield:              the name of the field
        :return                     if the field exists then a list is returned holding the appropriate entry
                                    if the field doesn't exist then an empty list is returned
        """
    return mk_optional_basic_field(measurement_type, 0, data, jfield)


def mk_bounded_int(measurement_type, data, jfield):
    """
    create a (measurement_type, valtype, value for the jfield in the data) triple for a bounded int
    :param measurement_type: measurement type of jfield in the data warehouse
    :param data: json that may contain the jfield
    :param jfield: the name of the field
    :return: Error free, [(measurement_type, valtype, value for the jfield in the data)], error_message
    """
    return mk_basic_field(measurement_type, 7, data, jfield)


def mk_optional_bounded_int(measurement_type, data, jfield):
    """
    If the jfield exists in the data then return [(measurement_ttpe, valtype, value for the jfield in the data)].
    If not then return an empty list.
        :param measurement_type:    measurement type of jfield in the data warehouse
        :param data:                json that may contain the jfield
        :param jfield:              the name of the field
        :return                     if the field exists then a list is returned holding the appropriate entry
                                    if the field doesn't exist then an empty list is returned
        """
    return mk_optional_basic_field(measurement_type, 7, data, jfield)


def mk_real(measurement_type, data, jfield):
    """
    create a (measurement_type, valtype, value for the jfield in the data) triple for a real
    :param measurement_type: measurement type of jfield in the data warehouse
    :param data: json that may contain the jfield
    :param jfield: the name of the field
    :return: Error free, [(measurement_type, valtype, value for the jfield in the data)], error_message
    """
    return mk_basic_field(measurement_type, 1, data, jfield)


def mk_optional_real(measurement_type, data, jfield):
    """
    If the jfield exists in the data then return [(measurement_ttpe, valtype, value for the jfield in the data)].
    If not then return an empty list.
        :param measurement_type:    measurement type of jfield in the data warehouse
        :param data:                json that may contain the jfield
        :param jfield:              the name of the field
        :return                     if the field exists then a list is returned holding the appropriate entry
                                    if the field doesn't exist then an empty list is returned
        """
    return mk_optional_basic_field(measurement_type, 1, data, jfield)


def mk_bounded_real(measurement_type, data, jfield):
    """
    create a (measurement_type, valtype, value for the jfield in the data) triple for a bounded int
    :param measurement_type: measurement type of jfield in the data warehouse
    :param data: json that may contain the jfield
    :param jfield: the name of the field
    :return: Error free, [(measurement_type, valtype, value for the jfield in the data)], error_message
    """
    return mk_basic_field(measurement_type, 8, data, jfield)


def mk_optional_bounded_real(measurement_type, data, jfield):
    """
    If the jfield exists in the data then return [(measurement_ttpe, valtype, value for the jfield in the data)].
    If not then return an empty list.
        :param measurement_type:    measurement type of jfield in the data warehouse
        :param data:                json that may contain the jfield
        :param jfield:              the name of the field
        :return                     if the field exists then a list is returned holding the appropriate entry
                                    if the field doesn't exist then an empty list is returned
        """
    return mk_optional_basic_field(measurement_type, 8, data, jfield)


def mk_string(measurement_type, data, jfield):
    """
    create a (measurement_type, valtype, value for the jfield in the data) triple for a string
    :param measurement_type: measurement type of jfield in the data warehouse
    :param data: json that may contain the jfield
    :param jfield: the name of the field
    :return: Error free, [(measurement_type, valtype, value for the jfield in the data)], error_message
    """
    return mk_basic_field(measurement_type, 2, data, jfield)


def mk_optional_string(measurement_type, data, jfield):
    """
    If the jfield exists in the data then return [(measurement_type, valtype, value for the jfield in the data)].
    If not then return an empty list.
    :param measurement_type:    measurement type of jfield in the data warehouse
    :param data:   json that may contain the jfield
    :param jfield: the name of the field
    :return if the field exists then a list is returned holding the appropriate entry
            if the field doesn't exist then an empty list is returned
    """
    return mk_optional_basic_field(measurement_type, 2, data, jfield)


def mk_unidecoded_string(measurement_type, data, jfield):
    """
    create a (measurement_type, valtype, value for the jfield in the data) triple for a string - remove accents
    :param measurement_type: measurement type of jfield in the data warehouse
    :param data: json that may contain the jfield
    :param jfield: the name of the field
    :return: Error free, [(measurement_type, valtype, value for the jfield in the data)], error_message
    """
    error_free, result, error_message = mk_basic_field(measurement_type, 2, data, jfield)
    if error_free:   # a result value exists
        measurement_type, val_type, val = result[0]  # pick the result triple from the list
        unidecoded_val = unidecode.unidecode(val)
        return error_free, [(measurement_type, val_type, unidecoded_val)], error_message
    else:
        return error_free, result, error_message


def mk_optional_unidecoded_string(measurement_type, data, jfield):
    """
    If the jfield exists in the data then return [(measurement_type, valtype, value for the jfield in the data)]
    Remove accents
    If not then return an empty list.
    :param measurement_type:    measurement type of jfield in the data warehouse
    :param data:   json that may contain the jfield
    :param jfield: the name of the field
    :return if the field exists then a list is returned holding the appropriate entry
            if the field doesn't exist then an empty list is returned
    """
    error_free, result, error_message = mk_optional_basic_field(measurement_type, 2, data, jfield)
    if error_free and (len(result) == 1):   # There is no error, and the optional string exists
        measurement_type, val_type, val = result[0]  # pick the result triple from the list
        unidecoded_val = unidecode.unidecode(val)
        return error_free, [(measurement_type, val_type, unidecoded_val)], error_message
    else:
        return error_free, result, error_message


def mk_datetime(measurement_type, data, jfield):
    """
    create a (measurement_type, valtype, value for the jfield in the data) triple for a datetime represented as a string
    :param measurement_type: measurement type of jfield in the data warehouse
    :param data: json that may contain the jfield
    :param jfield: the name of the field
    :return: Error free, [(measurement_type, valtype, value for the jfield in the data)], error_message
    """
    return mk_basic_field(measurement_type, 3, data, jfield)


def mk_optional_datetime(measurement_type, data, jfield):
    """
    Used for an optional datetime represented as a string
    If the jfield exists in the data then return [(measurement_type, valtype, value for the jfield in the data)].
    If not then return an empty list.
    :param measurement_type: measurement type of jfield in the data warehouse
    :param data: json that may contain the jfield
    :param jfield: the name of the field
    :return if the field exists then a list is returned holding the appropriate entry
            if the field doesn't exist then an empty list is returned
    """
    return mk_optional_basic_field(measurement_type, 3, data, jfield)


def get_and_check_datetime_from_epoch_ms(data, jfield):
    """
    check if a value exists, and if so convert it from an epoch (in ms) to a datetime string
    :param data: json that contains the jfield
    :param jfield: the name of the field
    :return: (field exists, well typed, value)
    """
    val = data.get(jfield)
    if val is None:
        exists = False
        well_typed = False  # default
        return exists, well_typed, ""
    elif val == "":
        exists = False
        well_typed = False  # default
        return exists, well_typed, ""
    else:
        exists = True
        well_typed, str_val = convert_epoch_in_ms_to_string(val)
        return exists, well_typed, str_val


def mk_datetime_from_epoch_in_ms(measurement_type, data, jfield):
    """
    create a (measurement_type, valtype, value for the jfield in the data) triple from epoch time in ms
    :param measurement_type: measurement type of jfield in the data warehouse
    :param data: json that may contain the jfield
    :param jfield: the name of the field
    :return: Error free, [(measurement_type, valtype, value for the jfield in the data)], error_message
    """
    date_time_type = 3
    (exists, well_formed, val) = get_and_check_datetime_from_epoch_ms(data, jfield)
    if exists and well_formed:
        return True, [(measurement_type, date_time_type, val)], ""
    else:
        if not exists:
            return False, [], missing_mandatory_type_error_message(jfield, measurement_type, data)
        else:  # must exist but not be well-formed
            return False, [], wrong_type_error_message(jfield, measurement_type, data, date_time_type)


def get_and_check_datetime_from_posix_timestamp(data, jfield):
    """
    check if a value exists, and if so convert it from a posix timestamp to a datetime string
    :param data: json that contains the jfield
    :param jfield: the name of the field
    :return: (field exists, well typed, value)
    """
    val = data.get(jfield)
    if val is None:
        exists = False
        well_typed = False  # default
        return exists, well_typed, ""
    elif val == "":
        exists = False
        well_typed = False  # default
        return exists, well_typed, ""
    else:
        exists = True
        well_typed, str_val = convert_posix_timestamp_to_string(val)
        return exists, well_typed, str_val


def mk_datetime_from_posix_timestamp(measurement_type, data, jfield):
    """
    Transform a CSV Posix timestamp field held in a string to a date/time format
    :param measurement_type:    measurement type of jfield in the data warehouse
    :param data: json that contains the jfield
    :param jfield: the name of the field
    :return: string formatted in year-month-day hrs:mins:sec
    """
    date_time_type = 3
    (exists, well_formed, val) = get_and_check_datetime_from_posix_timestamp(data, jfield)
    if exists and well_formed:
        return True, [(measurement_type, date_time_type, val)], ""
    else:
        if not exists:
            return False, [], missing_mandatory_type_error_message(jfield, measurement_type, data)
        else:  # must exist but not be well-formed
            return False, [],  wrong_type_error_message(jfield, measurement_type, data, date_time_type)


def mk_optional_datetime_from_epoch_in_ms(measurement_type, data, jfield):
    """
    Used when from epoch time in ms
    If the jfield exists in the data then return [(measurement_type, valtype, value for the jfield in the data)].
    If not then return an empty list.
    :param measurement_type:    measurement type of jfield in the data warehouse
    :param data:   json that may contain the jfield
    :param jfield: the name of the field
    :return if the field exists then a list is returned holding the appropriate entry
            if the field doesn't exist then an empty list is returned
    """
    date_time_type = 3
    (exists, well_formed, val) = get_and_check_datetime_from_epoch_ms(data, jfield)
    if exists and well_formed:
        return True, [(measurement_type, date_time_type, val)], ""  # jfield is present in the json and no errors
    elif exists and not well_formed:  # jfield is present but there are errors
        return False, [], wrong_type_error_message(jfield, measurement_type, data, date_time_type)
    else:
        return True, [], ""  # Field doesn't exist, which is OK as this is an optional field


def mk_boolean(measurement_type, data, jfield):
    """
    Transform a boolean represented as a 'T' or 'F' string into boolean field with 0 for False or 1 for True
    :param measurement_type:    measurement type of jfield in the data warehouse
    :param data:   json that may contain the jfield
    :param jfield: the name of the field
    :return : Error free, [(measurement_type, valtype, value for the jfield in the data)], error_message
    """
    (exists, well_formed, val, error_messsage) = get_and_check_value(measurement_type, 4, data, jfield, False)
    # val_type is set to 2 for checking as the field is expected to be a string ("T" or "Y") or ("F" or "N")
    if exists and well_formed:
        if val in ['0', 'N', 'F', 0]:
            val01 = mk_bool_string(False)
        else:
            val01 = mk_bool_string(True)  # must be 'Y' or 'T' or '1' or 1
        return True, [(measurement_type, 4, val01)], ""
    else:
        return False, [], error_messsage


def mk_optional_boolean(measurement_type, data, jfield):
    """
    If the jfield exists in the data then return [(measurement_ttpe, valtype, value for the jfield in the data)].
    If not then return an empty list.
        :param measurement_type:    measurement type of jfield in the data warehouse
        :param data:                json that may contain the jfield
        :param jfield:              the name of the field
        :return                     if the field exists then a list is returned holding the appropriate entry
                                    if the field doesn't exist then an empty list is returned
        """
    (exists, well_formed, val, error_message) = get_and_check_value(measurement_type, 2, data, jfield, True)
    # val_type is set to 2 for checking as the field is expected to be a string "T" or "F"
    if exists and well_formed:
        if val in ['0', 'N', 'F', 0]:
            val01 = mk_bool_string(False)
        else:
            val01 = mk_bool_string(True)  # must be 'Y' or 'T' or '1' or 1
        return True, [(measurement_type, 4, val01)], ""
    elif exists and not well_formed:
        return False, [], error_message
    else:
        return True, [], ""  # it's optional so OK if the field is not found


def mk_category_from_dict(cat_name, cat_dict, measurement_type):
    """
    Returns the category id from the category name.
    :param cat_name: the category name from the category table
    :param cat_dict: a directory with the category names as keys, and the categoryid as the values
    :param measurement_type: the measurement type (only used for debugging)
    :return Error free, [(measurement_type, valtype, value for the jfield in the data)], error_message
    """
    val = cat_dict.get(cat_name)
    if val is None:
        return False, val, f'\"{cat_name}\" is not in the dictionary for measurement type {measurement_type}'
    else:
        return True, val, ""


def mk_category_field(measurement_type, val_type, data, jfield, cat_dict):
    """
    make a category triple (measurement_type, value type, value)
    :param measurement_type: the id of the measurementtype that holds the value
    :param val_type: the type of the value to be stored: it is either 5 (nominal) or 6 (ordinal)
    :param data: the json structure
    :param jfield: the name of the field
    :param cat_dict: the dictionary for this category
    :return: Error free, [(measurement_type, valtype, value for the jfield in the data)], error_message
    """
    #  use val_type of 2 as string expected
    (exists, well_formed, val, error_message) = get_and_check_value(measurement_type, 2, data, jfield, False)
    if exists and well_formed:
        (ok, cat, category_error_message) = mk_category_from_dict(val, cat_dict, measurement_type)
        if ok:
            return True, [(measurement_type, val_type, cat)], ""
        else:
            return False, [], category_error_message
    else:
        return False, [], error_message


def mk_optional_category_field(measurement_type, val_type, data, jfield, cat_dict):
    """
    If the jfield exists in the data then return [(measurement_ttpe, valtype, value for the jfield in the data)].
    If not then return an empty list.
    :param measurement_type: the id of the measurementtype that holds the value if it exists
    :param val_type: the type of the value to be stored: it is either 5 (nominal) or 6 (ordinal)
    :param data: the json structure
    :param jfield: the name of the optional field
    :param cat_dict: the dictionary for this category
    :return: if the field exists then a list is returned holding the appropriate entry
             if the field doesn't exist then an empty list is returned
    """
    #  use val_type of 2 as string expected
    (exists, well_formed, val, error_message) = get_and_check_value(measurement_type, 2, data, jfield, True)
    if exists and well_formed:
        (ok, cat, category_error_message) = mk_category_from_dict(val, cat_dict, measurement_type)
        if ok:
            return True, [(measurement_type, val_type, cat)], ""
        else:
            return False, [], category_error_message
    elif exists and not well_formed:
        return False, [], error_message
    else:
        return True, [], ""  # it's optional, so it's OK that the field is missing


def mk_nominal(measurement_type, data, jfield, cat_dict):
    """
    make a nominal triple (measurement_type, 5, value)
    :param measurement_type: the id of the measurementtype that will hold the value
    :param data: the json structure
    :param jfield: the name of the field
    :param cat_dict: the dictionary for this category
    :return: Error free, [(measurement_type, valtype, value for the jfield in the data)], error_message
    """
    return mk_category_field(measurement_type, 5, data, jfield, cat_dict)


def mk_ordinal(measurement_type, data, jfield, cat_dict):
    """
    make an ordinal triple (measurement_type, 6, value)
    :param measurement_type: the id of the measurementtype that will hold the value
    :param data: the json structure
    :param jfield: the name of the field
    :param cat_dict: the dictionary for this category
    :return: Error free, [(measurement_type, valtype, value for the jfield in the data)], error_message
    """
    return mk_category_field(measurement_type, 6, data, jfield, cat_dict)


def mk_nominal_from_id(measurement_type, data, jfield):
    """
    make a nominal triple (measurement_type, 5, value) where the id is stored in the field
    :param measurement_type: the id of the measurementtype that will hold the value
    :param data: the json structure
    :param jfield: the name of the field
    :return: Error free, [(measurement_type, valtype, value for the jfield in the data)], error_message
    """
    return mk_basic_field(measurement_type, 5, data, jfield)


def mk_ordinal_from_id(measurement_type, data, jfield):
    """
    make an ordinal triple (measurement_type, 6, value) where the id is stored in the field
    :param measurement_type: the id of the measurementtype that will hold the value
    :param data: the json structure
    :param jfield: the name of the field
    :return: Error free, [(measurement_type, valtype, value for the jfield in the data)], error_message
    """
    return mk_basic_field(measurement_type, 6, data, jfield)


def mk_optional_nominal_from_dict(measurement_type, data, jfield, cat_dict):
    """
    If the jfield exists in the data then return [(measurement_ttpe, valtype, value for the jfield in the data)].
    If not then return an empty list.
    :param measurement_type: the id of the measurementtype that hold the value if it exists
    :param data: the json structure
    :param jfield: the name of the optional field
    :param cat_dict: the dictionary for this category
    :return: if the field exists then a list is returned holding the appropriate entry
             if the field doesn't exist then an empty list is returned
    """
    return mk_optional_category_field(measurement_type, 5, data, jfield, cat_dict)


def mk_optional_nominal_from_id(measurement_type, data, jfield):
    """
    make a nominal triple (measurement_type, 6, value) where the id is stored in the field if the field isn't empty
    :param measurement_type: the id of the measurementtype that will hold the value
    :param data: the json structure
    :param jfield: the name of the field
    :return: if the field exists then a list is returned holding the appropriate entry;
             if the field doesn't exist then an empty list is returned
    """
    return mk_optional_basic_field(measurement_type, 5, data, jfield)


def mk_optional_ordinal_from_dict(measurement_type, data, jfield, cat_dict):
    """
    If the jfield exists in the data then return [(measurement_ttpe, valtype, value for the jfield in the data)].
    If not then return an empty list.
    :param measurement_type: the id of the measurementtype that holds the value if it exists
    :param data: the json structure
    :param jfield: the name of the optional field
    :param cat_dict: the dictionary for this category
    :return: if the field exists then a list is returned holding the appropriate entry
             if the field doesn't exist then an empty list is returned
    """
    return mk_optional_category_field(measurement_type, 6, data, jfield, cat_dict)


def mk_optional_ordinal_from_id(measurement_type, data, jfield):
    """
    make an ordinal triple (measurement_type, 5, value) where the id is stored in the field if the field isn't empty
    :param measurement_type: the id of the measurementtype that will hold the value
    :param data: the json structure
    :param jfield: the name of the field
    :return: if the field exists then a list is returned holding the appropriate entry;
             if the field doesn't exist then an empty list is returned
    """
    return mk_optional_basic_field(measurement_type, 6, data, jfield)


def mk_bool_string(bool_val):
    """
    represent a boolean as a string ready to be stored in the db as an integer value
    :param bool_val: True or False
    :return: '0' or '1'
    """
    if bool_val:
        return '1'
    else:
        return '0'


def split_enum(measurement_types, data, jfield, valuelist):
    """
    ENUMS (Sets of values) are not represented directly in the warehouse. Instead they are represented as one boolean
    measurementtype per value. This function takes a json list of values and creates the list of measurements from it -
    one for each type.
    :param measurement_types: the list of measurement_types of the measurements into which the booleans are to be stored
    :param data: the json structure
    :param jfield: the name of the field
    :param valuelist: the list of values in the ENUM (in the order in which they are to be added into the
                    measurement types
    :return The list of (measurement_type,valType,value) triples that are used by
                    insertMeasurementGroup to add the measurements
    """
    res = []
    values = data.get(jfield)  # try to read the the list of values
    if values is None:
        exists = False
    elif values == "":
        exists = False
    else:
        exists = True
    if exists:
        for (measurement_type, value) in zip(measurement_types, valuelist):
            res = res + [(measurement_type, 4, mk_bool_string(value in values))]  # the 4 is because the type is boolean
        return True, res, ""
    else:
        return False, [], f'Missing Mandatory ENUM field {jfield} for measurement_types {measurement_types} in {data}'


def split_optional_enum(measurement_types, data, jfield, valuelist):
    """
    ENUMS (Sets of values) are not represented directly in the warehouse. Instead they are represented as one boolean
    measurementtype per value. This function takes a json list of values and creates the list of measurements from it -
    one for each type.
    :param measurement_types: the list of measurement_types of the measurements into which the booleans are to be stored
    :param data: the json structure
    :param jfield: the name of the field
    :param valuelist: the list of values in the ENUM (in the order in which they are to be added into the
                    measurement types
    :return The list of (measurement_type,valType,value) triples that are used by
                    insertMeasurementGroup to add the measurements
    """
    res = []
    values = data.get(jfield)  # try to read the the list of values
    if values is None:
        exists = False
    elif values == "":
        exists = False
    else:
        exists = True
    if exists:
        for (measurement_type, value) in zip(measurement_types, valuelist):
            res = res + [(measurement_type, 4, mk_bool_string(value in values))]  # the 4 is because the type is boolean
        return True, res, ""
    else:
        return True, [], ""  # Field doesn't exist, which is OK as this is an optional field


def get_converter_fn(event_type, mapper_dict):
    """
    map from the event_type to the mapper function and message group
    :param event_type: the e-Science Central event type
    :param mapper_dict: the dictionary that maps from the event_type to the mapper function and message group
    :return: (boolean indicating if the event_type is found, mapper function, measurement group
    """
    found = event_type in mapper_dict
    if found:
        info = mapper_dict[event_type]
        res = found, info["fn"], info["mg"]
    else:
        res = (found, None, None)
    return res
